//////////////////////////////////////////////////////////////
// Btree.cpp: Btree class methods
// Copyright(c) 1993 Azarona Software. All rights reserved.
//////////////////////////////////////////////////////////////
#include <string.h>
#include "btree.h"

Btree::Btree(int cs)
// Create a btree with a null file manager, and with a 
// cache size of cs
: f(0), cache(cs), root(cache)
{
}

/////////////////////////////////////
// File management functions
/////////////////////////////////////

int Btree::Connect(FmgrPtr &fptr, int create, long bha)
// Connect to an already open file. If create is 1, we're
// creating a new btree (but not necessarily a new file.)
// Returns SUCCESS or FAIL.
{
  f = fptr; bh_addr = bha;
  cache.Connect(f);
  if (create) {
     new(root) Mnode;
     bh.root_addr = root;
     bh.height = 1; bh.num_entries = 0; 
     bh.num_nodes = 1;  bh.order = NBR;
     WriteHdr();
  }
  else {
     ReadHdr();
     root = bh.root_addr;
  }
  if (f->OK()) return SUCCESS; else return FAIL;
}

void Btree::Disconnect()
// Disconnects the btree from the file.
{
  if (f) {
     WriteHdr();       
     root.Release();     // So we won't have dangling ptrs
     cache.Disconnect(); // Disconnect cache from the file
     f = 0;              // And ditto with btree
  }
}

int Btree::Create(char *fname, long bha)
// Creates a new file to hold the btree. Disconnects from any 
// file that we may be connected to first. Returns SUCCESS 
// or FAIL.
{
  Disconnect();
  FmgrPtr tmp(new Fmgr);
  if (tmp == 0) return FAIL;
  tmp->Create(fname, sizeof(BtreeHeader));
  if (!tmp->OK()) return FAIL;
  return Connect(tmp, 1, bha);
}

int Btree::Open(char *fname, Fmgr::AccessMode mode, long bha)
// Opens an existing file to hold the btree. Disconnect
// from any file that we may be connected to first.
// Returns SUCCESS or FAIL.
{
  Disconnect();
  FmgrPtr tmp(new Fmgr);
  if (tmp == 0) return FAIL;
  tmp->Open(fname, mode);
  if (!tmp->OK()) return FAIL;
  return Connect(tmp, 0, bha);
}

void Btree::WriteHdr()
{
  if (!f->ReadOnly())
     f->Store(&bh, sizeof(BtreeHeader), bh_addr);
}

void Btree::ReadHdr()
{
  f->Fetch(&bh, sizeof(BtreeHeader), bh_addr);
}

void Btree::Flush(int clear)
// Flush all buffers to disk
{
  if (f->OK()) {
     // When flushing the cache in read-only mode, no data
     // is actually written. However, we call Flush() any-
     // way to check for dangling coptrs.
     cache.Flush(clear);
     if (f->ReadyForWriting()) {
        WriteHdr();
        f->Flush();
     }
  }
}

////////////////////////////////////////////////
// Searching routines
////////////////////////////////////////////////

int Btree::Search(Entry &e)
// Search the tree for the first node having a matching
// entry (ie: keys must match). If found, the data
// field of e is filled in. Returns SUCCESS or FAIL.
{
  MnodePtr t(root);
  int rv, posn;

  while(t) { 
    rv = t->Search(e, posn); // Searching node t
    if (rv == 0) {           // Found a match
       e = t->entry[posn];
       return SUCCESS;
    }
    t = t->Branch(posn);     // No match, keep walking
  }

  return FAIL;
}


int Btree::FullSearch(const Entry &e)
// Like Search(), except both keys and data must match.
// Returns SUCCESS or FAIL.
{
  MnodePtr t(root);
  int rv, posn;

  while(t) {
    rv = t->FullSearch(e, posn);  // Searching node t
    if (rv == 0) {                // Found a full match
       return SUCCESS;
    }
    t = t->Branch(posn);          // No match, keep walking
  }

  return FAIL;
}

///////////////////////////////////////////////////////////////////
// Insertion routines
///////////////////////////////////////////////////////////////////

int Btree::Insert(Entry &e, MnodePtr &t)
// Recursive function that tries to insert entry e into subtree
// t. Returns SUCCESS, DUPLKEY, ALLOCERR, or NODE_OVERFLOW. 
// If NODE_OVERFLOW, then e becomes the median_entry to pass 
// back to t's parent.
{
  int rv, posn;

  if (t == 0) {
     // We went beyond the leaves. Pretend that we've overflowed
     e.right = 0;
     return NODE_OVERFLOW;
  }

  // Search the node t. If entry found, send back duplicate key error

  rv = t->FullSearch(e, posn);
  if (rv == 0) return DUPLKEY;

  // Entry not found, walk down to the appropriate leaf to store
  // entry.

  MnodePtr child(t);
  child = t->Branch(posn);
  t.Release(); // Prevents cache from filling up
  rv = Insert(e, child);

  // Back from recursion, see if we have an overflow

  child.Release(); // Prevents cache from filling up
  if (rv != NODE_OVERFLOW) return rv;

  // We have overflowed the node below us. See if we can
  // insert the entry e (now the median) into node t.

  posn++; // Move to the appropriate entry for insertion

  if (!t->IsFull()) { // Simply add t to node t
     t->InsEntry(e, posn);
     t->SetDirty();
     return SUCCESS;
  }
  else { // Need to split node t.

    // Create new node to be the right node. t will be the left node.

    MnodePtr r(t);
    new(r) Mnode;
    if (r == 0) return ALLOCERR;
    bh.num_nodes++;

    // Move roughly half the nodes to the right node. If inserting
    // entry into left node, we need to move the split position back
    // one, so that when the median is taken, the split node will have
    // enough entries.

    int split_posn = NBR/2;
    if (posn < split_posn) split_posn--;

    // BUG FIX: The following statements are due to a BC++ 
    // 3.1 error, that occurs when you turn inline functions
    // on. What we want to say is "t->Split(*r, split_posn).

    MnodeBucket &r_ref = *r;
    t->Split(r_ref, split_posn);

    t->SetDirty();

    // Figure out where entry e should go

    if (posn > split_posn) { // e goes to the right node
       // First entry of right node is the median
       Entry median_entry = r->entry[0];
       r->DelEntry(0);
       // Insert e into it's rightful spot
       r->InsEntry(e, posn-split_posn-1);
       e = median_entry; // For passing up the tree
    }
    // BUG FIX: Added the case posn == 0 below.
    else if (posn < split_posn || posn == 0) { // e goes to the left node
       t->InsEntry(e, posn);  t->SetDirty();
       e = r->entry[0]; // Get median from the right
       r->DelEntry(0);
    }

    // At this point, e is median entry to add to parent. 
    // Record what e's right pointer is going to be.
    r->left = e.right;
    e.right = r;

    return NODE_OVERFLOW;
  }
}

int Btree::Add(char *k, long d)
// Creates a new entry with key k and data a, and attempts to
// add the entry to the tree. Returns SUCCESS, DUPLKEY, or
// ALLOCERR.
{
  Entry e(k, d);
  int rv = Insert(e, root);
  if  (rv == NODE_OVERFLOW) { // Need to grow a new root
      MnodePtr old_root(root);
      new(root) Mnode;
      if (root == 0) return ALLOCERR;
      bh.num_nodes++; bh.height++;
      bh.root_addr = root;
      root->left = old_root;
      root->InsEntry(e, 0);
      rv = SUCCESS;
  }
  if (rv == SUCCESS) bh.num_entries++;
  return rv;
}

////////////////////////////////////////////////////////////////////
// Deletion routines
////////////////////////////////////////////////////////////////////

int Btree::Remove(char *k, long d)
// Deletes entry having key k and data d from the tree. 
// Returns SUCCESS, or FAIL if we couldn't find the entry
{
  Entry e(k, d);
  int rv = Delete(e, root);
  if (rv == SUCCESS && root->IsEmpty() && bh.height > 1)  {
     // Finish last merging operation which shortens the tree.
     // We don't let the tree go completely empty.
     MnodePtr p = root;
     root = root->Branch(-1);
     bh.root_addr = root;
     bh.num_nodes--; bh.height--;
     p.Delete();
  }
  return rv;
}

int Btree::Delete(Entry &e, MnodePtr &p)
// Recursive function that deletes entry e from the subtree
// with root p. Returns SUCCESS, or FAIL if we couldn't find
// the entry.
{
  MnodePtr t(p), s(p);
  Entry se; // Must be here or compiler generates the wrong code
  int rv, posn, sr;

  sr = p->FullSearch(e, posn); // Search node p
  t = p->Branch(posn);         // Get ready to walk down the tree

  if (sr == 0) {
     // p has entry to delete. Replace the entry with its
     // successor, if there is one.
     if (t) {
        s = t;
        while(s->Branch(-1)) s = s->Branch(-1);
        p->entry[posn] = s->entry[0];
        p->Branch(posn) = t; // Remember to restore the branch
        p->SetDirty();       // Don't forget
        // Now, using recursion, delete the successor entry
        // Release the coptrs in use, to keep cache from
        // filling up during the recursion, and to prevent
        // dangling pointer exceptions
        p.Release();
        se = s->entry[0];
        s.Release();
        if (Delete(se, t) != SUCCESS)
           excp->THROW(ASSERTERR); // We should have found the entry
     }
     else { // At a leaf, so the deletion is easy
        p->DelEntry(posn);
        p->SetDirty(); // Don't forget
        bh.num_entries--;
     }
     rv = SUCCESS;
  }
  else {
     if (t) {
        // Walk down the tree, keep looking for entry
        p.Release(); // Keep cache from filling up on recursive call
        rv = Delete(e, t);
     }
     else return FAIL; // Couldn't find said entry
  }

  if (rv == SUCCESS && t && t->IsPoor()) {
     // The node below p doesn't have enough entries. We need to
     // move some entries to it and thereby restore the balance.
     t.Release(); // To prevent dangling ptrs
     RestoreBalance(p, posn);
  }

  return rv;
}


void Btree::RestoreBalance(MnodePtr &p, int posn)
// Node down branch at position posn in node p has one too
// few entries. Give it an entry from either its left or right
// sibling, or perhaps just merge the node with a sibling.
{
  MnodePtr t(p);

  if (posn == -1) {
     // We have no left sibling, so try the right sibling
     t = p->Branch(0);
     if (t->IsPlentiful()) { // Can we borrow from the right?
        RotateLeft(p, 0);    // Do so.
     }
     else { // Right sibling has no spare entries.
       t.Release(); // To prevent dangling ptr errors
       Merge(p, 0); // Merge with right node and parent entry
     }
  }
  else if (posn == p->LastPosn()) {
     // We have no right sibling, so try the left sibling
     t = p->Branch(posn-1);
     if (t->IsPlentiful()) {  // Can we borrow from the left?
        RotateRight(p, posn); // Do so.
     }
     else { // Left sibling has no spare entries.
        t.Release();    // To prevent dangling ptr errors
        Merge(p, posn); // Merge with left node and parent entry
     }
  }
  else {
    // We have both left and right siblings
    t = p->Branch(posn-1);
    if (t->IsPlentiful()) {  // Can we borrow from the left?
       RotateRight(p, posn); // Do so.
    }
    else { 
       t = p->Branch(posn+1);
       if (t->IsPlentiful()) {   // Can we borrow from the right?
          RotateLeft(p, posn+1); // Do so.
       }
       else {
          // Neither the left or right sibling has spare entries.
          // Merge arbitrarily with the left node.
          t.Release();    // To prevent dangling ptr errors
          Merge(p, posn); // Merging with the left
       }
    }
  }
}

void Btree::RotateRight(MnodePtr &p, int posn)
// Does a "right rotation" using the entry at node p,
// position posn as the pivot point. Assumes p is not
// a leaf and that there is a left and right child.
// Also assumes right child isn't full, and that p and
// left child aren't empty.
{
  MnodePtr l(p), r(p);

  r = p->Branch(posn);   // Right child of p
  l = p->Branch(posn-1); // Left child of p

  // Move entry from parent p into rightmost entry of right node.
  // This entry will have the old left branch of right node. (The
  // left branch will be updated later.)

  p->Branch(posn) = r->left;       p->SetDirty();
  r->InsEntry(p->entry[posn], 0);  r->SetDirty();

  // Now, move rightmost entry of the left node into p. This
  // entry's branch becomes the right node's left pointer.
  // Be sure to point entry to right node.

  int last_posn = l->LastPosn();
  r->left = l->Branch(last_posn);       r->SetDirty();
  l->Branch(last_posn) = r;             l->SetDirty();
  p->entry[posn] = l->entry[last_posn]; p->SetDirty();
  l->DelEntry(last_posn);               l->SetDirty();

}

void Btree::RotateLeft(MnodePtr &p, int posn)
// Does a "left rotation" using the entry at node p,
// position posn as the pivot point. Assumes p is not
// a leaf and that there is a left and right child.
// Also assumes left child isn't full, and that p and
// right child aren't empty.
{
  MnodePtr l(p), r(p);

  r = p->Branch(posn);   // Right child of p
  l = p->Branch(posn-1); // Left child of p

  // Move entry from parent p into leftmost entry of left node.
  // This entry gets the left pointer of the right node

  p->Branch(posn) = r->left;        p->SetDirty();
  l->Concatenate(p->entry[posn]);   l->SetDirty();

  // Now, move rightmost entry of the right node into p. Make
  // sure this entry points to the right node. Also, we have
  // a new left branch of the right node.

  r->left = r->Branch(0);  
  r->Branch(0) = r;                r->SetDirty();
  p->entry[posn] = r->entry[0];    p->SetDirty();
  r->DelEntry(0);                  r->SetDirty();

}

void Btree::Merge(MnodePtr &p, int posn)
// Merges the node on the branch left of the entry at position
// posn of node p, with the entry of p and the node on the 
// branch to the right of the entry of p. Assumes posn in range.
{
  MnodePtr l(p), r(p); // Ptrs to left and right children

  r = p->Branch(posn);    // Right child of p
  l = p->Branch(posn-1);  // Left child of p

  // (1) Fixup entry of p to point to left branch of r
  // (2) Insert this into leftmost entry of l
  // (3) Add all of r's entries
  // (4) Delete entry from p, then delete r

  p->Branch(posn) = r->left;        p->SetDirty();
  l->Concatenate(p->entry[posn]);   l->SetDirty();

  // BUG FIX:
  // The following fixes an error in BC++ 3.1 when you
  // have inline functions turned on. What we want to
  // say is "l->Concatenate(*r);    l->SetDirty();

  MnodeBucket &r_ref = *r;
  l->Concatenate(r_ref);            l->SetDirty();

  p->DelEntry(posn);                p->SetDirty();
  r.Delete();
  bh.num_nodes--;
}


/////////////////////////////////////////////////////////
// Print routines
/////////////////////////////////////////////////////////

void Btree::PrintNode(MnodePtr &n)
// For testing purposes
{
  int i = 0;
  long addr = n;
  printf("a:%ld  ", addr);
  while (i < n->cnt) {
    printf("<");
    printf("%s, ", n->entry[i].key);
    printf("%ld", n->entry[i].data);
    printf("> ");
    i++;
  }
  printf("\n");
}

void Btree::PrintTree(MnodePtr &t, int sp)
// For testing purposes
{
  int i;

  if (t) {
     for(i = 0; i<sp; i++) putchar(' ');
     PrintNode(t);
     int n = t->cnt;
     MnodePtr p(t);
     for(i = -1; i<n; i++) {
        p = t->Branch(i);
        t.Release();
        PrintTree(p, sp+2);  
     }
  }
}

void Btree::Statistics(int full)
// For testing purposes.
{
  printf("Tree stats: %u entries, ", bh.num_entries);
  printf("%u nodes, %d levels\n", bh.num_nodes, bh.height);
  Report(cache, full);
}
